import asyncio
import aiohttp_cors
import aiohttp_jinja2
import jinja2
from aiohttp import web
from collections import defaultdict
from .auth import (is_authenticated, register,
                   generate_token, remove_token,
                   get_user_from_token)
import datetime
from .core import Attempt


db = None
workspace = None
wrapper_map = None
epoch = datetime.datetime.utcfromtimestamp(0)


async def check_auth(data, db):
    tok = data.get('token')
    if not (await is_authenticated(tok, db)):
        raise web.HTTPNotFound(reason='Not Logged in')


@aiohttp_jinja2.template('home.html')
async def home(request):
    questions = []
    async for q in db.questions.find().sort('qno'):
        questions.append(q['qid'])
    return {"questions": questions}


async def question(request):
    data = await request.json()
    await check_auth(data, db)
    qid = data.get('qid')
    q = await db.questions.find_one({"qid": qid},
                                    projection={"_id": False,
                                                'test_cases': False})
    if q is None:
        raise web.HTTPNotFound(reason='No such Question')
    return web.json_response({"statement": q['statement']})


async def new_attempt(request):
    data = await request.json()
    await check_auth(data, db)
    qid = data.get('qid')
    code = data.get('code')
    lang = data.get('lang')
    user_token = data.get('token')
    user = await get_user_from_token(user_token, db)
    wrap = wrapper_map.get(lang)
    attempt = Attempt(code, wrap, workspace, user, qid)
    await db.attempt_queue.insert_one(attempt.__dict__)
    return web.json_response({'attid': attempt.attid})


async def languages(request):
    return web.json_response({"languages": list(wrapper_map.keys())})


async def signup(request):
    data = await request.json()
    uname = data.get('uname')
    pwd = data.get('pwd')
    reg_ok, reason = await register(uname, pwd, db)
    if not reg_ok:
        raise web.HTTPNotFound(reason=reason)
    else:
        return web.json_response({})


async def login(request):
    data = await request.json()
    uname, pwd = data.get('uname'), data.get('pwd')
    tok_ok, tok = await generate_token(uname, pwd, db)
    if not tok_ok:
        raise web.HTTPNotFound(reason=tok)
    else:
        return web.json_response({"token": tok})


async def logout(request):
    data = await request.json()
    tok = data.get('token')
    await remove_token(tok, db)
    return web.json_response({})


async def score_calc(request):
    data = await request.json()
    uname = await get_user_from_token(data.get('token'), db)
    score = 0
    async for att in db.history.find({"user": uname}):
        score += 1 if att['status'] else 0
    return web.json_response({"score": score})


async def wait_for_verdict(request):
    data = await request.json()
    await check_auth(data, db)
    attid = data['attid']
    while True:
        att = await db.history.find_one({"attid": attid})
        if att is None:
            asyncio.sleep(1)
        else:
            return web.json_response({'status': att['status']})


async def leaderboard(request):
    history = defaultdict(list)
    async for attempt in db.history.find({"status": True}).sort('datetime'):
        att = Attempt()
        att.__dict__ = attempt
        stamp = att.datetime - epoch
        history[att.user].append((stamp.total_seconds(), 1))
    return web.json_response(history)


def run_server(port, host, database, static_folder,
               wrapmap, wkspace, template_path):
    global db, workspace, wrapper_map
    db = database
    workspace = wkspace
    wrapper_map = wrapmap
    app = web.Application()

    app.router.add_post('/login', login)
    app.router.add_post('/logout', logout)
    app.router.add_post('/signup', signup)
    app.router.add_post('/attempt', new_attempt)
    app.router.add_post('/question', question)
    app.router.add_get('/languages', languages)
    app.router.add_post('/score', score_calc)
    app.router.add_post('/wait/for/verdict', wait_for_verdict)
    app.router.add_get('/leader', leaderboard)
    app.router.add_get('/', home)
    # -----------cors

    corsconfig = {"*": aiohttp_cors.ResourceOptions(allow_credentials=True,
                                                    expose_headers="*",
                                                    allow_headers="*")}
    cors = aiohttp_cors.setup(app, defaults=corsconfig)
    for route in list(app.router.routes()):
        try:
            cors.add(route)
        except Exception as e:
            print(e)  # /register will be added twice and will raise error

    aiohttp_jinja2.setup(app, loader=jinja2.FileSystemLoader(template_path))

    x = app.router.add_static('/static', static_folder)
    x = cors.add(x)
    web.run_app(app, host=host, port=port)



import os
import random


def random_string(n=15):
    return ''.join([random.choice('abcdefghijklmnopqrstuvwxyz')
                    for _ in range(n)])


def normalize(string):
    return string.lower().strip()


async def add_questions_from_dir(dirname, time_limit, db):
    questions = [(i, os.path.join(dirname, i))
                 for i in os.listdir(dirname)]
    for qno, qpath in questions:
        qno = int(qno)
        q = await db.questions.find_one({"qno": qno})
        if q is None:
            print('Adding a new question', qno)
            q = {"qno": qno, "qid": random_string()}
            with open(os.path.join(qpath, 'statement'), 'r') as fl:
                q['statement'] = fl.read()
            ios = set([i[1:] for i in os.listdir(qpath) if i[0] in 'io'])
            q['test_cases'] = []
            for no in ios:
                tcase = {}
                with open(os.path.join(qpath, 'i'+no), 'r') as fl:
                    tcase['inp'] = fl.read()
                with open(os.path.join(qpath, 'o'+no), 'r') as fl:
                    tcase['out'] = fl.read()
                tcase['time_limit'] = time_limit
                q['test_cases'].append(tcase)
            await db.questions.insert_one(q)



from subprocess import (run, PIPE, TimeoutExpired)
from .utils import random_string, normalize
from datetime import datetime
import os


class TestCase:
    """
    This is an individual test case.


        t = TestCase(i, o, 10)
        status, out, err = t(code)
    """
    def __init__(self, inp, out, time_limit):
        self.inp = inp
        self.out = out
        self.time_limit = time_limit

    def __call__(self, attempt):
        """Run command on this input and produce verdict"""
        ipath = os.path.join(attempt.workspace, random_string())
        with open(ipath, 'w') as fl:
            fl.write(self.inp)
        cmd = attempt.wrapper.format(input_file=ipath,
                                     code_file=attempt.codepath)
        try:
            p = run(cmd, timeout=self.time_limit,
                    stderr=PIPE, stdout=PIPE,
                    shell=True)
        except TimeoutExpired:
            result = False
            stdout, stderr = b'', b'Timeout'
        else:
            stdout, stderr = p.stdout, p.stderr
            result = p.returncode == 0
            result = result and stdout.decode().strip() == self.out.strip()
        return result, stdout.decode(), stderr.decode()


class Question:
    """
    A single question in some contest.

        q = Question(statement, tests)
        status, log = q(attempt)
    """
    def __init__(self, qno=None, statement=None, test_cases=None):
        self.statement = statement
        self.test_cases = test_cases
        self.qno = qno
        self.qid = random_string()

    def __call__(self, attempt):
        """Run this attempt on this question"""
        log = [test(attempt) for test in self.test_cases]
        status = all([i[0] for i in log])
        attempt.status = status
        attempt.log = log
        return attempt


class Attempt:
    """
    This defines an attempt on some question
    """
    def __init__(self, code=None, wrapper=None, workspace=None,
                 user=None, qid=None):
        """
            a = Attempt(code, wrap, './work/abc')


            code        : The string of the program which it needs to work
            wrap        : The wrapper of the program which makes it work
            workspace   : The common workspace where this attempt is stored
        """
        self.code = code
        self.user = normalize(user) if user is not None else None
        self.qid = qid
        self.wrapper = wrapper
        self.attid = random_string(100)
        self.workspace = None
        self.codepath = None
        self.datetime = datetime.utcnow()
        if workspace is not None:
            wk = os.path.join(workspace, random_string())
            while os.path.exists(wk):
                wk = os.path.join(workspace, random_string())
            self.workspace = wk
            os.mkdir(self.workspace)
            self.codepath = os.path.join(self.workspace, random_string())
            with open(self.codepath, 'w') as fl:
                fl.write(self.code)



import os
import json
import asyncio
import argparse
from .server import run_server
from motor import motor_asyncio
from .utils import add_questions_from_dir
from .worker import run_judge
from .default_interface import copy_defaults


def get_db(uri):
    db_name = uri.split('/')[-1]
    client = motor_asyncio.AsyncIOMotorClient(uri)
    db = client[db_name]
    return db


def _add_questions(qdir, timeout, db):
    loop = asyncio.get_event_loop()
    todo = [asyncio.ensure_future(add_questions_from_dir(qdir, timeout, db))]
    loop.run_until_complete(asyncio.gather(*todo))


def main():
    mongo_uri = os.environ.get("MONGO_URI",
                               "mongodb://localhost:27017/openjudge")
    desc = 'Voody Ecommerce server library'
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('--server-host', action='store',
                        default='0.0.0.0',
                        help='What host to run the server on?')
    parser.add_argument('--server-port', action='store',
                        default=8000,
                        help='What port to run the server on?')
    parser.add_argument("--mongo-uri", action="store",
                        default=mongo_uri,
                        help='The MongoDB URI')
    parser.add_argument('--workspace', action='store',
                        default='workspace',
                        help='Folder to contain all working data')
    parser.add_argument('--template-dir', action='store',
                        default='templates',
                        help='Where are the templates stored?')
    parser.add_argument('--static-dir', action='store',
                        default='staticfiles',
                        help='Where are the static files stored?')
    parser.add_argument('--wrapmap', action='store',
                        default='wrappers.json',
                        help='Wrappers in JSON format')
    parser.add_argument('--add-questions-from', action='store',
                        default='questions',
                        help="Where to add questions from?")
    parser.add_argument('--timeout', action='store',
                        default=10,
                        help="How much total time does each attempt get?")
    parser.add_argument('--judge', action='store_true',
                        default=False,
                        help='Start the judge')
    parser.add_argument('--n-judges', action='store',
                        default=4, help='How many concurrent judges')
    args = parser.parse_args()
    copy_defaults(args.template_dir, args.static_dir, args.wrapmap)
    database = get_db(args.mongo_uri)
    if not os.path.exists(args.workspace):
        os.mkdir(args.workspace)

    with open(args.wrapmap, 'r') as fl:
        wrapmap = json.load(fl)

    if args.judge:
        run_judge(args.mongo_uri, args.n_judges)

    _add_questions(args.add_questions_from, args.timeout, database)
    # ------------------------------------------
    print('Initiating Server')
    run_server(args.server_port, args.server_host,
               database,
               args.static_dir,
               wrapmap,
               args.workspace,
               args.template_dir)
